// Demo.cpp : demonstrate how to use the library of raid-subsystem
//
#include <algorithm>
#include <vector>
#include <iostream>

// the most important header file for RAID controllers
#include "arclib.h"	

// header for SAS HBAs
#include "arclib-sas.h"

#include "ethernet.h"

using namespace std;
// includes the necessary interface header files
#if defined(_WIN32)
	#include "win_comm.h"
	#include "win_pass_ioctl.h"
	#include "win_scsi.h"
    #include "win_ioctl_sas.h"
#elif defined(__linux__) || defined(__FreeBSD__)
	#include "linux_comm.h"
	#include "linux_ioctl.h"
	#if defined(__linux__)
		#include "linux_scsi.h"
		#include "linux_pass_ioctl.h"
	#else
		#include "freebsd_scsi.h"
	#endif
#elif defined(__sun__)
	#include "solaris_comm.h"
	#include "solaris_ioctl.h"
#elif defined(__APPLE__)
	#include "darwin_userclient.h"
#endif


#if (_MSC_VER >= 1900) // vs2015
	#ifdef _WIN64
		#ifdef _DEBUG
			#pragma comment( lib, "..\\..\\bin\\windows-vs2015\\x86_64\\debug\\arclib.lib" )
		#else
			#pragma comment( lib, "..\\..\\bin\\windows-vs2015\\x86_64\\release\\arclib.lib" )
		#endif
	#else
		#ifdef _DEBUG
			#pragma comment( lib, "..\\..\\bin\\windows-vs2015\\i386\\debug\\arclib.lib" )		
		#else
			#pragma comment( lib, "..\\..\\bin\\windows-vs2015\\i386\\release\\arclib.lib" )	
		#endif
	#endif
#elif (_MSC_VER >= 1800) // vs2013
	#ifdef _WIN64
		#ifdef _DEBUG
			#pragma comment( lib, "..\\..\\bin\\windows-vs2013\\x86_64\\debug\\arclib.lib" )
		#else
			#pragma comment( lib, "..\\..\\bin\\windows-vs2013\\x86_64\\release\\arclib.lib" )
		#endif
	#else
		#ifdef _DEBUG
			#pragma comment( lib, "..\\..\\bin\\windows-vs2013\\i386\\debug\\arclib.lib" )		
		#else
			#pragma comment( lib, "..\\..\\bin\\windows-vs2013\\i386\\release\\arclib.lib" )	
		#endif
	#endif
#elif (_MSC_VER >= 1600) // vs2010
	#ifdef _WIN64
		#ifdef _DEBUG
			#pragma comment( lib, "..\\..\\bin\\windows\\x86_64\\debug\\arclib.lib" )
		#else
			#pragma comment( lib, "..\\..\\bin\\windows\\x86_64\\release\\arclib.lib" )
		#endif
	#else
		#ifdef _DEBUG
			#pragma comment( lib, "..\\..\\bin\\windows\\i386\\debug\\arclib.lib" )		
		#else
			#pragma comment( lib, "..\\..\\bin\\windows\\i386\\release\\arclib.lib" )	
		#endif
	#endif

#elif (_MSC_VER == 1400) // vs2005
	#ifdef _DEBUG
		#pragma comment( lib, "..\\..\\bin\\win32-vs2005\\debug\\arclib.lib" )
	#else
		#pragma comment( lib, "..\\..\\bin\\win32-vs2005\\release\\arclib.lib" )
	#endif

#endif//


#define MAX_CONTROLLER_SUPPORTED	64
CArclib ar[MAX_CONTROLLER_SUPPORTED];

/* from demo_hba.cpp */
extern ARC_STATUS DemoSASHBA(void);

#if defined(_WIN32)

/************************************************************************************
	Routine Name:	DiscoveryCommDevice
	Function:		Discover the controller(s) connected to the RS-232 port
	Input:			None.
	Output:			None.
	Return Values:	1. Total controller(s) found.
	Note:			None.
************************************************************************************/
int DiscoveryCommDevice()
{
	int nTotalCtrl = 0;

	WinCommInterface *iface = new WinCommInterface();
	for ( int i = 0; i < 4 ; i++ )
	{
		if ( iface->init(i) )
		{
			ar[nTotalCtrl++].ArcInitSession(iface);
			if ( nTotalCtrl < MAX_CONTROLLER_SUPPORTED )
			{
				iface = new WinCommInterface();                
			}
		}
	}

	return nTotalCtrl;
}

/************************************************************************************
	Routine Name:	DiscoveryCardDevice
	Function:		Discover the controller(s) connected to the PCI, PCI-X, PCI-E bus
	Input:			None.
	Output:			None.
	Return Values:	1. Total controller(s) found
	Note:			None.
************************************************************************************/
int DiscoveryCardDevice()
{
	int nTotalCtrl = 0;

	WinPassIoctlInterface *iface = new WinPassIoctlInterface();
	for ( int i = 0; i < 16 ; i++ )
	{
		if ( iface->init(i) )
		{
			ar[nTotalCtrl++].ArcInitSession(iface);
			if ( nTotalCtrl >= MAX_CONTROLLER_SUPPORTED )
			{
				break;
			}

			iface = new WinPassIoctlInterface();                       
		}
	}

	return nTotalCtrl;
}

/************************************************************************************
	Routine Name:	DiscoveryInbandDevice
	Function:		Discover the controller(s) connected to the Inband scsi bus
	Input:			None.
	Output:			None.
	Return Values:	1. Total controller(s) found.
	Note:			None.
************************************************************************************/
int DiscoveryInbandDevice()
{
	int nTotalCtrl = 0;

	WinScsiInterface *iface = new WinScsiInterface();
	for ( int scsi = 0; scsi <= MAX_SCSI_DEV ; scsi++ )
	{
		for ( int ch = 0; ch <= MAX_SCSI_CH ; ch++ )
		{
			for ( int target = 0; target <= MAX_SCSI_TARGETID ; target++ )
			{
				if ( iface->init(scsi, ch, target) )
				{
					ar[nTotalCtrl++].ArcInitSession(iface);
					if ( nTotalCtrl < MAX_CONTROLLER_SUPPORTED )
					{
						iface = new WinScsiInterface();
						if ( !iface )
						{
							printf("Internal Error\n");
							return -1;
						}
					}
					else
					{
						goto quit;
					}
				}
			}
		}
	}


	// scsn physical drive for our in-band device
	iface = new WinScsiInterface();
	for ( int pd = 0; pd <= MAX_SCSI_DEV ; pd++ )
	{
		if ( iface->init_physical_drive(pd) )
		{
			ar[nTotalCtrl++].ArcInitSession(iface);
			if ( nTotalCtrl < MAX_CONTROLLER_SUPPORTED )
			{
				iface = new WinScsiInterface();
				if ( !iface )
				{
					printf("Internal Error\n");
					return -1;
				}
			}
			else
			{
				goto quit;
			}
		}
	}


	quit:
	return nTotalCtrl;
}

#endif	// end of defined(_WIN32)


#if defined(__linux__)

/************************************************************************************
	Routine Name:	DiscoveryCommDevice
	Function:		Discover the controller(s) connected to the RS-232 port
	Input:			None.
	Output:			None.
	Return Values:	1. Total controller(s) found.
	Note:			None.
************************************************************************************/
int DiscoveryCommDevice()
{
	int nTotalCtrl = 0;

	LinuxCommInterface *iface = new LinuxCommInterface();
	for ( int i = 0; i < 4 ; i++ )
	{
		if ( iface->init(i) )
		{
			ar[nTotalCtrl++].ArcInitSession(iface);
			if ( nTotalCtrl >= MAX_CONTROLLER_SUPPORTED )
			{
				break;
			}

			iface = new LinuxCommInterface();            
		}
	}

	return nTotalCtrl;
}

/************************************************************************************
	Routine Name:	DiscoveryCardDeviceMethod2
	Function:		Discover the controller(s) connected to the PCI, PCI-X, PCI-E bus
					through pass ioctl interface
	Input:			None.
	Output:			None.
	Return Values:	1. Total controller(s) found
************************************************************************************/
	#define MAX_SCAN_SG	256
int DiscoveryCardDeviceMethod2()
{
	int nTotalCtrl = 0;
	int sg = 0;
	char status;

	do
	{
		LinuxPassIoctlInterface  *iface = new LinuxPassIoctlInterface();

		status = (int)iface->init(sg);
		if ( status )
		{
			ar[nTotalCtrl++].ArcInitSession(iface);
		}
		else
		{
			delete iface;
		}
	}while ( ++sg <= MAX_SCAN_SG && nTotalCtrl < MAX_CONTROLLER_SUPPORTED );

	return nTotalCtrl;

}

/************************************************************************************
	Routine Name:	DiscoveryCardDeviceMethod1
	Function:		Discover the controller(s) connected to the PCI, PCI-X, PCI-E bus
					through /proc/devices
	Input:			None.
	Output:			None.
	Return Values:	1. Total controller(s) found
************************************************************************************/
int DiscoveryCardDeviceMethod1()
{
	int nTotalCtrl = 0;

	const char  areca_driver_name[] = "arcmsr";
	int         major_num = 0;
	int         minor_num = 0;

	char        major_num_str[4];
	char        driver_str[7];
	char        dev_name[33];
	struct stat stat_of_dev;
	int         controller_index = 0;
	mode_t      mode = 0666;

	FILE *devices_file = NULL;

	devices_file = fopen("/proc/devices", "r");
	if ( !devices_file )
	{
		return 0;
	}


	// get areca's major#
	while ( fscanf(devices_file, "%3s%32s", major_num_str, dev_name) != EOF )
	{
		major_num_str[sizeof(major_num_str)-1] = 0;
		dev_name[sizeof(dev_name)-1] = 0;

		// find areca's major#
		if ( !memcmp(areca_driver_name, dev_name, 6) )
		{
			major_num = atoi(major_num_str);
			break;
		}
	}

	fclose(devices_file);

	if ( !major_num )
	{
		return 0;
	}

	mode |= S_IFCHR;

	for ( minor_num = 0; minor_num < MAX_CONTROLLER_SUPPORTED; minor_num++ )
	{
		sprintf(dev_name, "/dev/arcmsr%d", minor_num);
		if ( !stat(dev_name, &stat_of_dev) )
		{
			// device name found, check to see if the major and minor were correct.
			if ( major_num != (int)major(stat_of_dev.st_rdev) || 
				 minor_num != (int)minor(stat_of_dev.st_rdev) )
			{
				// major and minor number are not match, delete the old device name
				unlink(dev_name);
			}
			else
			{
				continue;
			}
		}

		if ( mknod(dev_name, mode, makedev(major_num, minor_num)) )
		{
			return 0;
		}
	}

	// setup areca controller(s)
	for ( controller_index = 0; controller_index < MAX_CONTROLLER_SUPPORTED; controller_index++ )
	{
		LinuxIoctlInterface *iface = new LinuxIoctlInterface();
		if ( iface )
		{
			if ( iface->init(controller_index) )
			{
				// call ArcInitSession(...) member function first before any others member functions get call
				ar[controller_index].ArcInitSession(iface);

				// save the total controller(s) we found
				nTotalCtrl++;
			}
			else
			{
				delete iface;
				iface = NULL;
			} // if
		}// if        
	}// for

	return nTotalCtrl;

}

/************************************************************************************
	Routine Name:	DiscoveryCardDevice
	Function:		Discover the controller(s) connected to the PCI, PCI-X, PCI-E bus
	Input:			None.
	Output:			None.
	Return Values:	1. Total controller(s) found
	Note:			None.
************************************************************************************/
int DiscoveryCardDevice()
{
	int nTotalCtrl = 0;
	// since the driver can accept two interface for client application, 
	// so if we could not find controller with primary method, try secondary.
	// 1. primary method: regular "major number" format, we first find the major number for areca raid subsystem
	//                    and make node for it, and finally open it for operation.
	// 2. secondary method: pass-through ioctl(sg interface) 
	//
	nTotalCtrl = DiscoveryCardDeviceMethod1();
	if ( nTotalCtrl == 0 )
	{
		// 
		// If primary method could not find any controller(s), try to use secondary method
		// 
		nTotalCtrl = DiscoveryCardDeviceMethod2();
	}

	return nTotalCtrl;

}

/************************************************************************************
	Routine Name:	DiscoveryInbandDevice
	Function:		Discover the controller(s) connected to the Inband scsi bus
	Input:			None.
	Output:			None.
	Return Values:	1. Total controller(s) found
	Note:			None.
************************************************************************************/
int DiscoveryInbandDevice()
{
	int    sg = 0;
	char   status;
	int    nTotalCtrl = 0;

	do
	{
		LinuxSCSIInterface  *iface = new LinuxSCSIInterface();

		status = (char)iface->init(sg);
		if ( status )
		{
			ar[nTotalCtrl++].ArcInitSession(iface);
			break;
		}
		else
		{
			delete iface;
		}
	}while ( ++sg <= 15 && nTotalCtrl < MAX_CONTROLLER_SUPPORTED );

	return nTotalCtrl;
}


#endif	// end of defined(__linux__)



#if defined (__FreeBSD__)

/************************************************************************************
	Routine Name:	DiscoveryCommDevice
	Function:		Discover the controller(s) connected to the RS-232 port
	Input:			None.
	Output:			None.
	Return Values:	1. Total controller(s) found.
	Note:			None.
************************************************************************************/
int DiscoveryCommDevice()
{
	int nTotalCtrl = 0;

	LinuxCommInterface *iface = new LinuxCommInterface();
	for ( int i = 0; i < 4 ; i++ )
	{
		if ( iface->init(i) )
		{
			ar[nTotalCtrl++].ArcInitSession(iface);
			if ( nTotalCtrl >= MAX_CONTROLLER_SUPPORTED )
			{
				break;
			}

			iface = new LinuxCommInterface();            
		}
	}

	return nTotalCtrl;
}

/************************************************************************************
	Routine Name:	DiscoveryCardDevice
	Function:		Discover the controller(s) connected to the PCI, PCI-X, PCI-E bus
	Input:			None.
	Output:			None.
	Return Values:	1. Total controller(s) found
	Note:			None.
************************************************************************************/
int DiscoveryCardDevice()
{
	char buf[64];
	char rmstr[128];
	mode_t mode = 0666;
	dev_t dev;
	int     nNumofPciDevs = 0;
	int nTotalCtrl = 0;

	mode |= S_IFCHR;


	// remove all exist device node first
	for ( int i = 0; i < 4; i++ )
	{
		memset(buf, 0, sizeof(buf));
		sprintf(buf,"/dev/arcmsr%d", i);
		unlink(buf);
	}


	LinuxIoctlInterface *out = new LinuxIoctlInterface();

	for ( int mymajor = 255; mymajor > 161; mymajor-- )
	{
		for ( int myminor = 0; myminor < 4; myminor++ )
		{
			memset(buf, 0, sizeof(buf));
			sprintf(buf,"/dev/arcmsr%d", nTotalCtrl);
			dev = makedev(mymajor, myminor);
			mknod(buf, mode, dev);
			if ( out->init(nTotalCtrl) )
			{
				ar[nTotalCtrl++].ArcInitSession(out);
				out = new LinuxIoctlInterface();
			}
			else
			{
				unlink(buf);
			}
		}


		if ( nTotalCtrl )
		{
			break;
		}
	}

	return nTotalCtrl;
}



/************************************************************************************
	Routine Name:	DiscoveryInbandDevice
	Function:		Discover the controller(s) connected to the Inband scsi bus
	Input:			None.
	Output:			None.
	Return Values:	1. Total controller(s) found
	Note:			None.
************************************************************************************/
int DiscoveryInbandDevice()
{
	int pass = 0;
	char status;
	int nTotalCtrl = 0;

	do
	{
		FreeBSDSCSIInterface  *iface = new FreeBSDSCSIInterface();

		for ( int retry = 0; retry < 5; retry++ )
		{
			status = (int)iface->init(pass);
			if ( status )
				break;
		}
		if ( status )
		{
			ar[nTotalCtrl++].ArcInitSession(iface);
			break;
		}
		else
		{
			delete iface;
		}
	}while ( ++pass <= 15 && nTotalCtrl < MAX_CONTROLLER_SUPPORTED );


	return nTotalCtrl;
}


#endif	// end of defined (__FreeBSD__)



#if defined(__sun__)
/************************************************************************************
	Routine Name:	DiscoveryCommDevice
	Function:		Discover the controller(s) connected to the RS-232 port
	Input:			None.
	Output:			None.
	Return Values:	1. Total controller(s) found.
	Note:			None.
************************************************************************************/
int DiscoveryCommDevice()
{   
	int nTotalCtrl = 0;

	char status;
	SolarisCommInterface *iface = new SolarisCommInterface();
	for ( int i = 0; i < MAX_CONTROLLER_SUPPORTED ; i++ )
	{
		status = (char)iface->init(i);
		if ( status )
		{
			ar[nTotalCtrl++].ArcInitSession(iface);
			iface = new SolarisCommInterface();
		}
	}

	if ( iface )
	{
		delete iface;
		iface = NULL;
	}

	return nTotalCtrl;
}

/************************************************************************************
	Routine Name:	DiscoveryCardDevice
	Function:		Discover the controller(s) connected to the PCI, PCI-X, PCI-E bus
	Input:			None.
	Output:			None.
	Return Values:	1. Total controller(s) found
	Note:			None.
************************************************************************************/
int DiscoveryCardDevice()
{
	int nTotalCtrl = 0;
	int controller_index;

	for ( controller_index = 0; controller_index < MAX_CONTROLLER_SUPPORTED; controller_index++ )
	{
		SolarisIoctlInterface *iface = new SolarisIoctlInterface();
		if ( iface )
		{
			if ( iface->init(controller_index) )
			{
				ar[controller_index].ArcInitSession(iface);
				nTotalCtrl++;
			}
			else
			{
				delete iface;
				iface = NULL;
			} // if
		}// if
	}// for
	return nTotalCtrl;
}



/************************************************************************************
	Routine Name:	DiscoveryInbandDevice
	Function:		Discover the controller(s) connected to the Inband scsi bus
	Input:			None.
	Output:			None.
	Return Values:	1. Total controller(s) found
	Note:			None.
************************************************************************************/
int DiscoveryInbandDevice()
{
	// not implement yet...
	return 0;
}

#endif



#if defined(__APPLE__)
/************************************************************************************
Routine Name:	DiscoveryCommDevice
Function:		Discover the controller(s) connected to the RS-232 port
Input:			None.
Output:			None.
Return Values:	1. Total controller(s) found.
Note:			None.
************************************************************************************/
int DiscoveryCommDevice()
{
	// not implement yet...
	return 0;
}

/************************************************************************************
	Routine Name:	DiscoveryCardDevice
	Function:		Discover the controller(s) connected to the PCI, PCI-X, PCI-E bus
	Input:			None.
	Output:			None.
	Return Values:	1. Total controller(s) found
	Note:			None.
************************************************************************************/
int DiscoveryCardDevice()
{
	int nTotalCtrl = 0;
	int controller_index;

	for ( controller_index = 0; controller_index < MAX_CONTROLLER_SUPPORTED; controller_index++ )
	{
		DarwinUserClientInterface *iface = new DarwinUserClientInterface();

		if ( iface->init(controller_index) )
		{
			ar[controller_index].ArcInitSession(iface);
			nTotalCtrl++;
		}
		else
		{
			delete iface;
			iface = NULL;
		}
	}
	return nTotalCtrl;
}



/************************************************************************************
	Routine Name:	DiscoveryInbandDevice
	Function:		Discover the controller(s) connected to the Inband scsi bus
	Input:			None.
	Output:			None.
	Return Values:	1. Total controller(s) found
	Note:			None.
************************************************************************************/
int DiscoveryInbandDevice()
{
	// not implement yet...
	return 0;
}


#endif

/************************************************************************************
	Routine Name:	DiscoveryNetDevice
	Function:		Discover the controller(s) on the ethernet
	Input:			None.
	Output:			None.
	Return Values:	1. Total controller(s) found
	Note:			None.
************************************************************************************/
int DiscoveryNetDevice()
{
	ARC_STATUS  status;
	int total_ctrl = 0;

	// allocate an instance of the "Interface" and initialize it with a proper IP address
	EthernetInterface *ei = new EthernetInterface();
	if ( !ei->init("192.168.0.161") )
	{
		return total_ctrl;
	}

	// assign the interface to the library
	status = ar[0].ArcInitSession(ei);
	if ( status != ARC_SUCCESS )
	{
		delete ei;
		return total_ctrl;
	}

	// return just one controller, you can add additional controllers by instancing a EthernetInterface
	total_ctrl = 1;
	return total_ctrl;
}

/*****************************************************************
Function: Sub-routine of "Hardware Monitor Information" command
Input:	1.argc
		2.argv
Output:	1. ARC_STATUS
*****************************************************************/

ARC_STATUS Exe_Hw_Info( CArclib &ctrl )
{
	ARC_STATUS  status;
	sSYSTEM_INFO    sysInfo;
	sHwMonInfo  voltageInfo, fanInfo, hddTempInfo, batteryInfo;
	sHwMon5010  hwInfo5010;
	int     i;
	int		addTemp;
	bool    got5010 = false;
	bool    gotRaidCard = false;

	char strId[64];
	char modelStr[9];

	memset(strId, 0, sizeof(strId));
	memset(modelStr, 0, sizeof(modelStr));

	status = ctrl.ArcGetIdentifyString(strId);
	if ( status != ARC_SUCCESS )
	{
		return status;
	}

	memcpy(modelStr, (const void*)&strId[20], 8);


	if ( strcmp(modelStr, "ARC-5010")==0 || strcmp(modelStr, "ARC-6010")==0 )
	{
		got5010 = true;
	}

	if ( strncmp(modelStr, "ARC-11", 6) == 0 || strncmp(modelStr, "ARC-12", 6)==0 )
	{
		gotRaidCard = true;
	}

	if ( got5010 )
	{
		status = ctrl.ArcGetHWMon5010(&hwInfo5010);
		if ( status != ARC_SUCCESS )
		{
			return status;
		}
	}
	else
	{
		status = ctrl.ArcGetSysInfo(&sysInfo);
		if ( status != ARC_SUCCESS )
		{
			return status;
		}

		status = ctrl.ArcGetHWMon(HWMON_FAN_INFO, &fanInfo);
		if ( status != ARC_SUCCESS )
		{
			return status;
		}

		status = ctrl.ArcGetHWMon(HWMON_TMP_INFO, &hddTempInfo);
		if ( status != ARC_SUCCESS )
		{
			return status;
		}

		status = ctrl.ArcGetHWMon(HWMON_VOL_INFO, &voltageInfo);
		if ( status != ARC_SUCCESS )
		{
			return status;
		}

		status = ctrl.ArcGetHWMon(HWMON_BATTERY_INFO, &batteryInfo);
		if ( status != ARC_SUCCESS )
		{
			return status;
		}

	}

	printf("The Hardware Monitor Information\n");
	printf("===========================================\n");
	if ( got5010 )	  // 5 bays ARC-5010 or ARC-6010
	{
		printf("Controller Temp : %d\n", hwInfo5010.monCtrlTemp);
		printf("BackPlane Temp  : %d\n", hwInfo5010.monBkPlaneTemp);
		printf("Fan Speed (RPM) : %d\n", hwInfo5010.monFan);
		printf("Power +12.00V   : %2.3lf\n", (double)hwInfo5010.mon12V/(double)1000.0);
		printf("Power +5.00V    : %2.3lf\n", (double)hwInfo5010.mon5V/(double)1000.0);
		printf("Power +3.30V    : %2.3lf\n", (double)hwInfo5010.mon3_3V/(double)1000.0);
		printf("Power +1.50V    : %2.3lf\n", (double)hwInfo5010.mon1_5V/(double)1000.0);
	}
	else if ( gotRaidCard )	   // RaidCard
	{
		for ( i = 0; i < fanInfo.HwItemCount; i++ )
		{
			if ( fanInfo.ItemInfo[i] == 0xffff )
			{
				printf("Fan#%d Speed (RPM)  : N.A.\n", i+1);
			}
			else
			{
				printf("Fan#%d Speed (RPM)  : %d\n", i+1, fanInfo.ItemInfo[i]);
			}
		}

		for ( i = 0; i < batteryInfo.HwItemCount; i++ )
		{
			if ( batteryInfo.ItemInfo[i] == BBM_NOT_INSTALLED )
			{
				printf("Battery Status      : Not Installed\n");
			}
			else if ( batteryInfo.ItemInfo[i] == BBM_FAILED )
			{
				printf("Battery Status      : Failed\n");
			}
			else
			{
				printf("Battery Status      : %d%%\n", batteryInfo.ItemInfo[i]);
			}
		}

		// show temperatures 
		addTemp = 0;
		i = 0;

		if(hddTempInfo.HwItemCount > sysInfo.gsiIdeChannels)
		{
			printf("CPU Temp.           : %d\n", hddTempInfo.ItemInfo[0]);
			printf("Controller Temp.    : %d\n", hddTempInfo.ItemInfo[1]);
			addTemp = 2;
		}

		for (i = 0; addTemp < hddTempInfo.HwItemCount; i++, addTemp++)
		{
			printf("HDD #%-2d Temp.       : %d\n", i + 1, hddTempInfo.ItemInfo[addTemp]);
		}            
	}
	else // subsystem
	{
		for ( i = 0; i < voltageInfo.HwItemCount; i+=2 )
		{
			printf("Power +%-5.2lfV       : %2.3lf\n", (double)voltageInfo.ItemInfo[i]/(double)1000.0, 
				   (double)voltageInfo.ItemInfo[i+1]/(double)1000.0);
		}

		for ( i = 0; i < fanInfo.HwItemCount; i++ )
		{
			if ( fanInfo.ItemInfo[i] == 0xffff )
			{
				printf("Fan#%d Speed (RPM)  : N.A.\n", i+1);
			}
			else
			{
				printf("Fan#%d Speed (RPM)  : %d\n", i+1, fanInfo.ItemInfo[i]);
			}
		}

		// show controller temperature first
		printf("Controller Temp.    : %d\n", hddTempInfo.ItemInfo[0]);
		//printf("Battery Status     : Not Installed\n");
		for ( i = 1; i < hddTempInfo.HwItemCount; i++ )
		{
			printf("HDD #%-2d Temp.       : %d\n", i, hddTempInfo.ItemInfo[i]);
		}       

	}
	printf("===========================================\n");

	return ARC_SUCCESS;
}

ARC_STATUS Exe_SAS_Hw_Info( CArclib &ctrl )
{
	ARC_STATUS  status = ARC_SUCCESS;   
	BYTE max_enc = 0;
	sSYSTEM_INFO    sysInfo;
	memset(&sysInfo, 0, sizeof(sSYSTEM_INFO));
	status = ctrl.ArcGetSysInfo(&sysInfo);
	if ( status != ARC_SUCCESS )
	{
		return status;
	}
	max_enc = sysInfo.gsiNumberEnclosures;
	//
	// The controller itself h/w monitor information resides at (max enclosure index + 1)
	// ex: A raidset has max up to 8 enclosures, the enclosure index is from 0~7,
	//     then the controller h/w monitor information is index 8
	BYTE contrller_index = max_enc;    
	sSENSORS sensors;
	int loop;

	printf("The Hardware Monitor Information\n");
	printf("=====================================================\n");
	printf("[Controller H/W Monitor]\n");
	status = ctrl.ArcGetEnclosureHwMonitor(contrller_index, &sensors);
	if ( status == ARC_SUCCESS )
	{
		for ( loop = 0; loop < sensors.temperature.snNumberOfElements; loop++ )
		{
			printf("  %-16s : %ld C\n", 
				   sensors.temperature.snElementStr[loop], 
				   sensors.temperature.snValue[loop]);
		}

		for ( loop = 0; loop < sensors.fan.snNumberOfElements; loop++ )
		{
			printf("  %-16s : %ld RPM\n", 
				   sensors.fan.snElementStr[loop], 
				   sensors.fan.snValue[loop]);
		}

		for ( loop = 0; loop < sensors.voltage.snNumberOfElements; loop++ )
		{
			printf("  %-16s : %2.3lf V\n", 
				   sensors.voltage.snElementStr[loop], 
				   (double)(sensors.voltage.snValue[loop]) / 1000);
		}

		for ( loop = 0; loop < sensors.battery.snNumberOfElements; loop++ )
		{
			if ( sensors.battery.snValue[loop] == BBM_NOT_INSTALLED )
			{
				printf("  %-16s : Not Installed\n", sensors.battery.snElementStr[loop]);
			}
			else if ( sensors.battery.snValue[loop] == BBM_FAILED )
			{
				printf("  %-16s : Failed\n", sensors.battery.snElementStr[loop]);
			}
			else
			{
				printf("  %-16s : %d%%\n", sensors.battery.snElementStr[loop], sensors.battery.snValue[loop]);
			}
		}
	}

	sGUI_ENC_PROPERTY encpro;
	BYTE    prodname[17];
	BYTE    vendorname[9];
	BYTE    rev[5];

	for ( int index = 0; index < sysInfo.gsiNumberEnclosures; index++ )
	{
		status = ctrl.ArcGetEnclosureProperty(index, &encpro);
		if ( status != ARC_SUCCESS || encpro.gepEncType == ENC_TYPE_NOT_INSTALLED )
			continue;


		// enclosure description
		memset(prodname, 0, sizeof(prodname));
		memset(vendorname, 0, sizeof(vendorname));
		memset(rev, 0, sizeof(rev));
		memcpy(prodname, encpro.gepProduct, sizeof(encpro.gepProduct));
		memcpy(vendorname, encpro.gepVendor, sizeof(encpro.gepVendor));
		memcpy(rev, encpro.gepRev, sizeof(encpro.gepRev));

		printf("[Enclosure#%d : %s%s%s]\n", index+1, vendorname, prodname, rev);

		status = ctrl.ArcGetEnclosureHwMonitor(index, &sensors);
		if ( status == ARC_SUCCESS )
		{
			for ( loop = 0; loop < sensors.voltage.snNumberOfElements; loop++ )
			{
				printf("  %-16s : %2.3lf V\n", 
					   sensors.voltage.snElementStr[loop], 
					   (double)(sensors.voltage.snValue[loop]) / 1000);
			}
			for ( loop = 0; loop < sensors.fan.snNumberOfElements; loop++ )
			{
				printf("  %-16s : %ld RPM\n", 
					   sensors.fan.snElementStr[loop], 
					   sensors.fan.snValue[loop]);
			}

			for ( loop = 0; loop < sensors.power.snNumberOfElements; loop++ )
			{
				printf("  %-16s : %s\n", 
					   sensors.power.snElementStr[loop], 
					   sensors.power.snValue[loop] == 1 ? "OK" : "FAILED");
			}

			for ( loop = 0; loop < sensors.temperature.snNumberOfElements; loop++ )
			{
				if ( sensors.temperature.snValue[loop] == -20 )
				{
					printf("  %-16s : %s\n", 
						   sensors.temperature.snElementStr[loop], "N.A.");
				}
				else
				{
					printf("  %-16s : %ld C\n", 
						   sensors.temperature.snElementStr[loop], 
						   sensors.temperature.snValue[loop]);
				}
			}
		}
	}


	printf("=====================================================\n");  
	return status;
}

#define		UNKNOWN_EVT_STR	"Unknown Event"
/*****************************************************************
Function: Show the event(s) of raid-subsystem
Input:
Output:	1. ARC_STATUS
*****************************************************************/
ARC_STATUS Exe_Event_Info( CArclib &ctrl )
{
	ARC_STATUS  status = ARC_SUCCESS;
	char        ptr1[64];				// temp space of string		
	char        tmpStr[64];
	int         nCount = 0;				// total number of event objects we got	
	int         i;  
	char        timeBuf[64];
	int         nEventFlag = FLAG_ALL;	// the flag pass into ArcGetReqEventPage(...)
	int         nTmpEventType;
	sSYSTEM_INFO    sysInfo;

	//the time event occurred, defined in event.h
	pSYS_TIME evtTime = (pSYS_TIME)new sSYS_TIME;
	// event object defined in event.h
	pEVENT_DATA pEventObject = (pEVENT_DATA)new sEVENT_DATA;    

	*((DWORD *)evtTime) = 0L;
	memset(pEventObject, 0, sizeof(sEVENT_DATA));

	status = ctrl.ArcGetSysInfo(&sysInfo);
	if ( status != ARC_SUCCESS )
	{
		goto quit;
	}

	/* 
		We should call ArcSetTime() to make the system time correct
	*/
	status = ctrl.ArcSetTime();
	if ( status != ARC_SUCCESS )
	{
		goto quit;
	}
	/* 
		Call ArcGetReqEventPage function to get the required event page from raid subsystem.
		There are 4 event pages reside in raid subsystem and each contains 16 event objects.			
	*/
	status = ctrl.ArcGetReqEventPage(nEventFlag, &nCount);
	if ( status != ARC_SUCCESS )
	{
		goto quit;
	}
	if ( status == ARC_SUCCESS )
	{
		if ( nCount != 0 )
		{
			printf("Date-Time               Device              Event Type	       \n");
			printf("===============================================================================\n");
			for ( i = 0; i <= nCount-1; i++ )
			{
				/*
					Call ArcGetEventObject function to extract the individual event object from event page
				*/
				if ( ctrl.ArcGetEventObject( i, pEventObject ) )
				{
					memset(timeBuf, 0, sizeof(timeBuf));
					// If tmFlag == 1 ==> use YY/MM/DD/HH/MM/SS else use time tick
					*((DWORD *)evtTime) = pEventObject->evtTime;
					if ( *((DWORD *)evtTime) & 0x80000000 )
					{
						sprintf(timeBuf, "%04d-%02d-%02d  %02d:%02d:%02d", 
								evtTime->u.tmYear + 2000, 
								evtTime->u.tmMonth,
								evtTime->u.tmDate,
								evtTime->u.tmHour,
								evtTime->u.tmMinute,
								evtTime->u.tmSecond);
					}
					else
					{
						sprintf(timeBuf, "%d", evtTime->x.tmTick );
					}

					memset( ptr1, 0, sizeof(ptr1));
					memset( tmpStr, 0, sizeof(tmpStr));

					/* 
						Parse the event object we got
					*/
					switch ( (int)pEventObject->evtCategory )
					{
					/*
						The member "evtStr" of event object structure contains the string to indicate
						which raidset owns this event object
					*/
					case EVENT_RAIDSET:		// raidset event		
						nTmpEventType = pEventObject->evtType;
						memcpy(tmpStr, pEventObject->evtStr, 16);
						printf("%-24s", timeBuf);
						if ( nTmpEventType > RS_EVT_NO_EVENT )
						{
							printf("%-20s%s\n", tmpStr, UNKNOWN_EVT_STR);
						}
						else
						{
							printf("%-20s%s\n", tmpStr, htmRaidEvent[pEventObject->evtType]);
						}                                           
						break;

						/*
							The member "evtStr" of event object structure contains the string to indicate
							which volumeset owns this event object
						*/
					case EVENT_VOLUMESET: // volumeset event
						nTmpEventType = pEventObject->evtType;
						memcpy(tmpStr, pEventObject->evtStr, 16);
						printf("%-24s", timeBuf);
						if ( nTmpEventType >= VS_EVT_TOTALS )
						{
							printf("%-20s%s\n", tmpStr, UNKNOWN_EVT_STR);
						}
						else
						{
							printf("%-20s%s\n", tmpStr, htmVolEvent[pEventObject->evtType]);
						}                   
						break;

					case EVENT_DEVICE:	// device event
						nTmpEventType = pEventObject->evtType;
						if ( sysInfo.gsiTargetType == TARGET_SAS )
						{
							memcpy(tmpStr, pEventObject->evtStr, 16);
						}
						else
						{
							sprintf(tmpStr, "IDE Channel #%02d", pEventObject->evtChannel + 1);  
						}
						printf("%-24s", timeBuf);
						if ( nTmpEventType >= DEV_EVT_TOTALS )
						{
							printf("%-20s%s\n", tmpStr, UNKNOWN_EVT_STR);
						}
						else
						{
							printf("%-20s%s\n", tmpStr, htmDevEvent[pEventObject->evtType]);
						}                                                                       
						break;

					case EVENT_HOST:	// host event
						/* this is Host SCSI events */
						nTmpEventType = pEventObject->evtType;
						printf("%-24s", timeBuf);
						if ( sysInfo.gsiScsiHostChannels )
						{
							if ( nTmpEventType >= SCSI_EVT_TOTALS )
							{
								printf("%-20s%s\n",
									   eventCat[pEventObject->evtCategory],
									   UNKNOWN_EVT_STR);
							}
							else
							{
								printf("%-20s%s\n",
									   eventCat[pEventObject->evtCategory],
									   htmScsiHostEvent[pEventObject->evtType]);
							}
						}
						else
						{
							if ( nTmpEventType >= SCSI_EVT_TOTALS )
							{
								printf("%-20s%s\n",
									   eventCat[pEventObject->evtCategory],
									   UNKNOWN_EVT_STR);
							}
							else
							{
								printf("%-20s%s\n",
									   eventCat[pEventObject->evtCategory],
									   htmIdeHostEvent[pEventObject->evtType]);
							}
						}
						break;

					case EVENT_HW_MONITOR:	// hardware monitor event
						nTmpEventType = pEventObject->evtType;
						printf("%-24s", timeBuf);
						if ( nTmpEventType >= HW_EVT_TOTALS )
						{
							printf("%-20s%s\n",
								   eventCat[pEventObject->evtCategory],
								   UNKNOWN_EVT_STR);
						}
						else
						{
							printf("%-20s%s\n",
								   eventCat[pEventObject->evtCategory],
								   htmHwMonEvent[pEventObject->evtType]);
						}
						break;
					case EVENT_NEW_83782D:
						memcpy(tmpStr, pEventObject->evtStr, 16);
						printf("%-24s", timeBuf);
						if ( pEventObject->evtType >= GHM_TOTALS )
						{
							printf("%-20s%s\n", tmpStr, UNKNOWN_EVT_STR);
						}
						else
						{
							printf("%-20s%s\n", tmpStr, htmNewEventStr[pEventObject->evtType]);
						}
						break;
					case EVENT_NO_EVENT:
//							printf("		No Event\n");
					default:
						break;

					}// end switch
				}// end if
			}// end for
			printf("===============================================================================\n");
		}//end if
	}// end if

	quit:
	if ( pEventObject )
		delete pEventObject;
	if ( evtTime )
		delete evtTime;

	return status;
}

ARC_STATUS  demo_create_raidset()
{
	ARC_STATUS      status;
	sDRV_MAP        drvMap;			// the buffer to contain the devices that were found by the raid controller
	sDEVICE_LIST    devList;		// the device list to create raidset	
	DWORD           nDevMask = 0;	// devices index for creating raidset(Non-SAS)
	sGUI_PHY_DRV    *drvInfo;		// see arclib.h for details		
	sSYSTEM_INFO    sysInfo;		// see arclib.h for details		
	vector <int>    v1;				// devices index for creating raidset(SAS)


	memset(&sysInfo, 0, sizeof(sysInfo));

	// use the controller#0 for demonstrating
	status = ar[0].ArcGetSysInfo(&sysInfo);
	if ( status != ARC_SUCCESS )
	{
		printf("Failed to ArcGetSysInfo(...), Error Code = %X.\n", status);
		return 1;
	}

	// check password
	status = ar[0].ArcChkPassword("0000");
	if ( status != ARC_SUCCESS )
	{
		printf("Failed to ArcChkPassword(...), Error Code = %X.\n", status);
		return 1;
	}

	// device enumeration
	status = ar[0].ArcHelpBuildDeviceMap(&drvMap);
	if ( status != ARC_SUCCESS )
	{
		printf("Failed to ArcHelpBuildDeviceMap(...), Error Code = %X.\n", status);
		return status;
	}


	// For demonstrating we suppose creating raidset with device 0~3, 
	// modify this with the real configuration
	v1.push_back(0);
	v1.push_back(1);
	v1.push_back(2);
	v1.push_back(3);

	// check to see if drive is available for creating raidset
	for ( int index = 0; index < (int)v1.size(); index++ )
	{
		drvInfo = &drvMap.drv[v1[index]].drvInfo;
		if ( !(drvInfo->gpdDeviceState & DEV_AVAILABLE) )
		{
			printf("Drive %d Is Not Available\n", v1[index]);
			return ARC_FAILURE;
		}
		else if ( drvInfo->gpdRaidNumber != 0xff )
		{
			printf("Drive %d Is Not Free\n", v1[index]);
			return ARC_FAILURE;                     
		}
		else
		{
			// fill out the device list for creating raidset
			// if Non-SAS the devList.device_pair[index].nEncIndex is always set to 0
			devList.device_pair[index].nEncIndex = drvMap.drv[v1[index]].encindex;
			devList.device_pair[index].nDrvIndex = drvMap.drv[v1[index]].drvindex;          
		}
	}

	// create raidset with default raidset name
	devList.total = (BYTE)v1.size();
	status = ar[0].ArcCreateRaidSet((pDEVICE_LIST)&devList, NULL);
	if ( status != ARC_SUCCESS )
	{
		printf("Failed to ArcCreateRaidSet(...), Error Code = %X.\n", status);
		return status;
	}

	
	sVolumeSetCap   volCap;
	memset(&volCap, 0, sizeof(sVolumeSetCap));
	volCap.structs_size = sizeof(sVolumeSetCap);
	volCap.raid = 0;	 // create this volume inside the first RaidSet
	strcpy((char *)volCap.volname, "VolumeSet#1");
	volCap.capacity = 10.0;
	volCap.raidlevel = VOLUME_TYPE_RAID5;
	volCap.stripsize = STRIPE_SIZE_64K;
	volCap.channel = 0;
	volCap.id = 0;
	volCap.lun = 0;
	volCap.tag = ENABLE;
	volCap.cache = ENABLE;
	volCap.speed = SCSI_SPEED_160;
	volCap.quickinit = FOREGROUND_INIT;

	status = ar[0].ArcCreateVolSet(&volCap);
	if ( status != ARC_SUCCESS )
	{
		printf("Failed to ArcCreateVolSet(...), Error Code = %X.\n", status);
		return 1;
	}

	return status;
}


ARC_STATUS demo()
{
	ARC_STATUS      status;
	sSYSTEM_INFO    sysInfo;		// see arclib.h for details			
	vector <int>    v1;				// devices index for creating raidset(SAS)

	memset(&sysInfo, 0, sizeof(sysInfo));

	// use the controller#0 for demonstrating
	status = ar[0].ArcGetSysInfo(&sysInfo);
	if ( status != ARC_SUCCESS )
	{
		printf("Failed to ArcGetSysInfo(...), Error Code = %X.\n", status);
		return status;
	}

	// check password
	status = ar[0].ArcChkPassword("0000");
	if ( status != ARC_SUCCESS )
	{
		printf("Failed to ArcChkPassword(...), Error Code = %X.\n", status);
		return status;
	}

	if ( sysInfo.gsiTargetType == TARGET_SAS )
	{
		// print out the h/w monitor information for SAS controller
		status = Exe_SAS_Hw_Info(ar[0]);
	}
	else
	{
		// print out the h/w monitor information for Non-SAS controller
		status = Exe_Hw_Info(ar[0]);
	}

	if ( status != ARC_SUCCESS )
	{
		return status;
	}

	// print out the events of raid-subsystem
	status = Exe_Event_Info(ar[0]);
	if ( status != ARC_SUCCESS )
	{
		return status;
	}

	status = demo_create_raidset();

	return status;
}

int FindController()
{
	int total_ctrl = 0;
	int ctrl = 0;


#define ENABLE_NET_DEVICE      0
#define ENABLE_COMM_DEVICE     1
#define ENABLE_INBAND_DEVICE   0
#define ENABLE_CARD_DEVICE     1
	//
	// choice an interface you want to implement, and comment or delete the others
	// 

	/* For Ethernet Interface */    
#if ENABLE_NET_DEVICE
	// Discover the controller that can be accessed through ethernet interface
	ctrl = DiscoveryNetDevice();
	total_ctrl += ctrl;
#endif

	/* For RS232 Interface */
#if ENABLE_COMM_DEVICE
	// Discover the controller that can be accessed through COM port interface	
	ctrl = DiscoveryCommDevice();
	total_ctrl += ctrl;
#endif

	/* For Ioctl Interface */
#if ENABLE_CARD_DEVICE
	// Discover the controller that can be accessed through PCI,PCI-X and PCI-E interface
	ctrl = DiscoveryCardDevice();
	total_ctrl += ctrl;
#endif

	/* For Inband Interface */  
#if ENABLE_INBAND_DEVICE
	// Discover the controller that can be accessed through SCSI inband interface
	ctrl = DiscoveryInbandDevice();
	total_ctrl += ctrl;
#endif

	return total_ctrl;
}


ARC_STATUS DemoRAIDCtrl()
{
	int total_ctrl = 0;
	ARC_STATUS  status = ARC_FAILURE;	// see arclib.h for details
	sSYSTEM_INFO    sysInfo;			// see arclib.h for details	
	int retval = 1;

	memset(&sysInfo, 0, sizeof(sysInfo));

	total_ctrl = FindController();
	printf("%d controller(s) found.\n", total_ctrl);
	if ( total_ctrl )
	{		
		// use the controller#0 for demonstrating
		status = ar[0].ArcGetSysInfo(&sysInfo);
		if ( status == ARC_SUCCESS )
		{
			status = demo();
			if ( status == ARC_SUCCESS )
			{
				retval = 0;
			}
		}
	}

	return status;
}


int main(int argc, char* argv[])
{
	ARC_STATUS  status;	

	/* demo for how to use RAID Controllers Library */
	status = DemoRAIDCtrl();

	printf("\n\n");

	/* demo for how to use SAS HBAs Library */
	status = DemoSASHBA();

	/* IMPORTANT: DO NOT delete the instance of the "Interface" by yourself,
	   library will take care of that */

	return status;
}


